#include <stdlib.h>
#include <stdio.h>
#include "queen.h"

int verbose = 0 ;

/**********************/
/* Subtask management */
/**********************/

void print_subtask(FILE *fp,subtask_t *z) {
  fprintf(fp,"%i %i %i",z->id,z->n,z->depth) ;
  for (int k = 0 ; k < z->depth ; k++) {
    fprintf(fp," %i",z->t[k]) ;
  }
  fprintf(fp, "\n") ;
}

subtask_t *alloc_arg(int id,int n,int depth) {
  subtask_t *r = calloc(sizeof(*r),1) ;
  if (!r) { perror("calloc") ; exit(2) ; }
  r->id = id ;
  r->n = n ;
  r->depth = depth ;
  r->t = calloc(sizeof(r->t[0]),r->depth) ;
  if (!r->t) { perror("calloc") ; exit(2) ; }
  return r ;
}

void free_arg(subtask_t *z) {
  free(z->t) ;
  free(z) ;
}

subtask_t *copy_arg(subtask_t *z) {
  subtask_t *r = alloc_arg(z->id,z->n,z->depth) ;
  for (int k = 0 ; k < z->depth; k++)
    r->t[k] = z->t[k];
  return r ;
}

static void error_read(char *msg) {
  fprintf(stderr, "Bad checkpoint file %s: %s\n", "(stdin)", msg) ;
  exit(2) ;
}

subtask_t *read_subtask(FILE *fp) {  
  int id,n,depth ;  
  if (fscanf(fp, "%i", &id) != 1) error_read("id") ;
  if (fscanf(fp, "%i", &n) != 1) error_read("n") ;
  if (fscanf(fp, "%i", &depth) != 1) error_read("depth") ;
  subtask_t *r = alloc_arg(id,n,depth) ;
  /* Seen array */
  for (int k = 0 ; k < r->depth ; k++) {
    if (fscanf(fp, "%i", &r->t[k]) != 1) error_read("seen[k]") ;
  }
  return r ;
}

/*************/
/* Utilities */
/*************/

inline static base set_all(int n) {
  return ((base)1 << n) - 1 ;
}


/* Generator */
static int dogen(subtask_t *z,base cols,base left,base right,int d,emit_t emit)
{
  if (d < 0) {
    emit(z) ;
    z->id++ ;
    return 1 ;
  }
  int r = 0 ;
  base forbid = cols | left | right ;
  for (int k = 0 ; k < z->n ; k++) {
    base col = (base)1 << k ;
    if (!(col & forbid)) {
      z->t[z->depth-d-1] = k ;
      r += dogen(z,cols|col,(left|col) <<1,(right|col)>>1,d-1,emit) ;
    }
  }
  return r ;
}


int generate_subtasks(int n, int depth,emit_t emit) {
  if (depth < 2) {
    fprintf(stderr,"depth argument must be >= 2\n");
    exit(2);
  }
  int r = 0 ;
  subtask_t *z = alloc_arg(0,n,depth) ;
  base col1 = 1 ;
  for (int k1 = 0 ; k1 < n-1 ; k1++) {
    z->t[0] = k1 ;
    // Next row
    base left1 = col1 << 1, right1 = col1 >> 1 ;
    base col2 = col1 << 2 ;
    for (int k2 = k1+2 ; k2 < n ; k2++) {
      z->t[1] = k2 ;
      r +=
        dogen(z,col1|col2,(left1|col2) << 1,(right1|col2) >> 1,depth-3,emit);
      col2 <<= 1 ;
    }
    col1 <<= 1 ;
  }
  free_arg(z);
  return r ;
}

/*******/
/* Run */
/*******/
#ifdef PP
static int nn;
static int d;
static base board[NMAX];

void pp_row(FILE *fp,base x) {
  for (int k = 0 ; k < nn ; k++) {
    fputc(x & 0x1 ? 'X':'.',fp) ;
    x >>=1 ;
  }
  fputc('\n',fp) ;
}

void pp_board(FILE *fp) {
  fprintf(fp,"********\n") ;
  for (int k=0 ; k < nn; k++) {
    pp_row(fp,board[k]) ;
  }
}
#endif


#ifndef OPT
static count_t dorun(base cols,base left,base right,base full) {
  //  fprintf(stderr,"ENTER:\n") ;
  //  fprintf(stderr,"cols=") ; pp_row(stderr,cols) ;
  //  fprintf(stderr,"left=") ; pp_row(stderr,left) ;
  //  fprintf(stderr,"right=") ; pp_row(stderr,right) ;

  base allowed = ~(cols | left | right) & full ;
  if (!allowed) return 0 ;
  base msk = allowed ^ (allowed-1) ;
  base col = allowed & msk ;
#ifdef PP
  board[d] = col ;
#endif
  base ncols = cols|col ;
  if (ncols == full) {
#ifdef PP
    pp_board(stdout) ;
#endif
    return 1;
  }
#ifdef PP
  d++ ;
#endif
  count_t r = dorun(ncols,(col|left) << 1,(col|right) >> 1,full) ;
#ifdef PP
  d-- ;
#endif
  allowed &= ~msk ;
  while (allowed) {
    col = allowed & (-allowed) ;
#ifdef PP
    board[d] = col ; d++ ;
#endif
    ncols = cols|col ;
    base nleft = (left | col) <<  1 ;
    base nright = (right | col) >> 1 ;
    r += dorun(ncols,nleft,nright, full) ;
#ifdef PP
    d-- ;
#endif
    allowed ^= col ;
  }
  return r ;
}

#else

#define MAXN 31
static count_t dorun(base cols,base left,base right,base full) {
  count_t num = 0 ;
  int sp = 0 ;
  base scols[MAXN], sleft[MAXN], sright[MAXN], sallowed[MAXN];

  cols |= ~full ;
  scols[0] = cols ; sleft[0] = left ; sright[0] = right ;
  base allowed = ~(cols|left|right)  ;
  while (sp >= 0) {
    while (allowed) {
      base msk = allowed & -allowed ;
      base ncols = scols[sp] | msk ;
      base nleft = (sleft[sp] | msk) << 1 ;
      base nright = (sright[sp] | msk) >> 1 ;
      base nallowed = ~(ncols|nleft|nright) ;
#ifdef PP
  board[d] = msk ;
#endif
      allowed ^= msk ;
      num += ncols == -1 ;
#ifdef PP
      if (ncols == -1)  pp_board(stdout) ;
#endif
      if (nallowed) {
#ifndef PP
        if (allowed) {
#endif
          sallowed[sp++]  = allowed ;
#ifdef PP
          d++;
#endif
#ifndef PP
        }
#endif
        scols[sp] =  ncols ;
        sleft[sp] = nleft ;
        sright[sp] = nright ;
        allowed = nallowed ;
      }
    }
#ifdef PP
    d--;
#endif
    allowed = sallowed[--sp] ;

  }
  return num ; 
}
#endif

count_t run_subtask(subtask_t *z) {
#ifdef PP
  d = 0 ; nn = z->n ;
#endif
  base cols=0, left=0, right=0;
  for (int k = 0 ; k < z->depth; k++) {    
    base col = (base)1 << z->t[k] ;
#ifdef  PP
    board[d++] =  col ;
#endif
    cols |= col ;
    left |= col ; left <<= 1 ;
    right |= col ; right >>= 1 ;
  }
  base full = set_all(z->n) ;

  // Allow pre-call check in dorun
  if (cols == full) return 1 ;

  count_t r = dorun(cols,left,right,full) ;
  return r ;
}

count_t run_task(int n) {
  count_t r = 0 ;
  base full = set_all(n) ;
  for (int q0=0; q0<n-2; q0++) {
    for (int q1=q0+2; q1<n; q1++){
      base c0 = (base)1 << q0 ; 
      base c1 = (base)1 << q1 ; 
      base cols = c0 | c1 ;
      base left = c0 ; left <<=1 ; left |= c1 ; left <<= 1 ;
      base right = c0 ; right >>=1 ; right |= c1 ; right >>= 1 ;
      r +=  dorun(cols,left,right,full) ;
    }
  }
  return 2*r ;
}
